from pathlib import Path, PosixPath
import psutil

# Paths
MAIN_DIR = PosixPath("~/.local/share/abtranslate").expanduser()
PACKAGE_DIR = MAIN_DIR/"packages"

# Url's 
PACKAGES_INDEX = None

# Values
BATCH_SIZE = 32
DEFAULT_CT2_CONFIG = {  
                        "compute_type" : 'default', 
                        "inter_threads" : psutil.cpu_count(logical=False), 
                        "intra_threads" : 0, 
                        "max_queued_batches" : 0, 
                        "flash_attention" : False, 
                        "tensor_parallel" : False, 
                        "files" : None}

DEFAULT_CT2_TRANSLATION_CONFIG = {  "target_prefix": None, 
                                    "max_batch_size": 0, 
                                    "batch_type": 'examples', 
                                    "asynchronous": False, 
                                    "beam_size": 2, 
                                    "patience": 1, 
                                    "num_hypotheses": 1, 
                                    "length_penalty": 1, 
                                    "coverage_penalty": 0, 
                                    "repetition_penalty": 1, 
                                    "no_repeat_ngram_size": 0, 
                                    "disable_unk": False, 
                                    "suppress_sequences":  None, 
                                    "end_token": None, 
                                    "return_end_token": False, 
                                    "prefix_bias_beta": 0, 
                                    "max_input_length": 1024, 
                                    "max_decoding_length": 256, 
                                    "min_decoding_length": 1, 
                                    "use_vmap": False, 
                                    "return_scores": False, 
                                    "return_logits_vocab": False, 
                                    "return_attention": False, 
                                    "return_alternatives": False, 
                                    "min_alternative_expansion_prob": 0, 
                                    "sampling_topk": 1, 
                                    "sampling_topp": 1, 
                                    "sampling_temperature": 1, 
                                    "replace_unknowns": False, 
                                    "callback" : None}


import pandas as pd
import pytest

from pyspark.sql import SparkSession 

from abtranslate.translator.package import load_argostranslate_model
from abtranslate.translator.argos_translator import ArgosTranslator
from abtranslate.translation.spark_translation import translate_with_udf

model_path = "./tests/example/model_data.argosmodel"
package = load_argostranslate_model(model_path)
sample_dataset_path = "./tests/example/aircraft_defects_chinese.csv"
test_df = pd.read_csv(sample_dataset_path) 
translator = package.load_translator(optimized_config=True)

sample_text_zh = """发动机左侧在飞行前检查中发现有轻微的机油渗漏。技术人员确认渗漏区域位于涡轮壳体附近，并进行了初步清洁。随后安排进一步的检查以确定是否需要更换密封件。

在机务人员建议下，更换了左侧发动机的油封并重新紧固相关组件。完成维护后进行了地面试车，未再发现新的渗漏现象。飞机恢复运行状态，记录保留供后续检查使用。"""

# @pytest.fixture
# def argos_translator() -> ArgosTranslator:
#     package = load_argostranslate_model("./model_sample/model_data.argosmodel")
#     return package.load_translator()

def test_translation():
    translation = translator.translate(sample_text_zh)
    print("Translation result: ", translation)

def test_batch_translation():
    input_column = test_df["input-text"]
    translated_column = translator.translate_batch(input_column)
    print("Df translaton results: ", translated_column)

def test_udf_translation():
    spark = SparkSession.builder.getOrCreate()
    df = spark.read.csv(sample_dataset_path, header=True)
    df_translated = translate_with_udf(model_path, df, input_column_name="input-text", output_column_name="translated_text")
    df_translated.select("translated_text").show(truncate=False)



from zipfile import ZipFile
from pathlib import Path

from .logger import logger 

def extract_package(model_file: str, extract_dir: str) -> bool:
    """
    Safely extract model files with.
    
    Args:
        model_file: Path to the model archive file
        extract_dir: Directory to extract files to
        
    Returns:
        Path to package directory if extraction successful, False otherwise
    """
    try:
        with ZipFile(model_file, 'r') as zip_ref:
            # Validate zip file integrity
            zip_ref.testzip()
            
            # Get list of files to extract
            file_list = zip_ref.namelist()
            
            # Check if files already exist
            extract_path = Path(extract_dir)
            all_exist = all(
                (extract_path / name).exists() 
                for name in file_list   
            )
            
            if not all_exist:
                logger.info(f"Extracting model files to {extract_dir}")
                zip_ref.extractall(extract_dir)
            else:
                logger.info("Model files already exist, skipping extraction")
            
            package_name = file_list[0]
            return Path(extract_dir)/package_name
            
    except Exception as e:
        raise Exception(f"Model extraction failed: {e}")
        # return False


import stanza

class Sentencizer:
    def split_sentences(text:str) -> list[str]:
        raise NotImplementedError()    
        
class StanzaSentencizer:
    def __init__(self, model_dir, lang_code):
        self.model_dir = str(model_dir)
        self.lang_code = lang_code
        self.processor = None

    def lazy_load(self):
        
        if not self.processor:
            self.processor = stanza.Pipeline(
                lang=self.lang_code,
                dir=self.model_dir,
                processors="tokenize",
                use_gpu=False,
                logging_level="WARNING",
            )
        return self.processor
        
    
    def split_sentences(self, text:str) -> list[str]:
        processor = self.lazy_load()
        sbd = processor(text)
        return [sentence.text for sentence in sbd.sentences]


from abtranslate.utils.helper  import flatten_list, expand_to_sentence_level, get_structure, apply_structure, is_valid_text
from abtranslate.translator.package import load_argostranslate_model
package = load_argostranslate_model("./tests/example/model_data.argosmodel")

rows = [
    """发动机左侧在飞行前检查中发现有轻微的机油渗漏。技术人员确认渗漏区域位于涡轮壳体附近，并进行了初步清洁。随后安排进一步的检查以确定是否需要更换密封件。

员建议下，更换了左侧发动机的油封并重新紧固相关组件。完成维护后进行了地面试车，未再发现新的渗漏现象。飞机恢复运行状态，记录保留供后续检查使用。""",

    """起落架在落地过程中右侧收回速度明显低于正常值。飞行员报告指示灯延迟亮起约五秒钟，怀疑液压系统响应缓慢。机务团队在停机坪进行了初步排查。

检查结果显示液压油液位偏低，已完成补充。同时对起落架作动筒进行了润滑保养并进行了重复测试。经测试，起落架收放功能恢复正常，未发现进一步异常。"""
]

expanded_sentences = [
    [ # Row 1
        [ # Paragraph 1 in Row 1
            '发动机左侧在飞行前检查中发现有轻微的机油渗漏。', 
            '技术人员确认渗漏区域位于涡轮壳体附近，并进行了初步清洁。', 
            '随后安排进一步的检查以确定是否需要更换密封件。'
        ], 
        #[], (blank) got this if ignore_empty_line=True on fuunction`expand_to_sentence_level` 
        [ # Paragraph 2 in Row 1
            '在机务人员建议下，更换了左侧发动机的油封并重新紧固相关组件。', 
            '完成维护后进行了地面试车，未再发现新的渗漏现象。', 
            '飞机恢复运行状态，记录保留供后续检查使用。']
    ], 
    [ # Row 2
        [ # Paragraph 1 in Row 2
            '起落架在落地过程中右侧收回速度明显低于正常值。', 
            '飞行员报告指示灯延迟亮起约五秒钟，怀疑液压系统响应缓慢。', 
            '机务团队在停机坪进行了初步排查。'
        ], 
        #[],  (blank) got this if ignore_empty_line=True on fuunction`expand_to_sentence_level` 
        [ # Paragraph 2 in Row 2
            '检查结果显示液压油液位偏低，已完成补充。', 
            '同时对起落架作动筒进行了润滑保养并进行了重复测试。', 
            '经测试，起落架收放功能恢复正常，未发现进一步异常。'
        ]
    ]
]

structure = [[3, 3], [3, 3]]

flattened_sentences = [ '发动机左侧在飞行前检查中发现有轻微的机油渗漏。', 
                        '技术人员确认渗漏区域位于涡轮壳体附近，并进行了初步清洁。', 
                        '随后安排进一步的检查以确定是否需要更换密封件。', 
                        '在机务人员建议下，更换了左侧发动机的油封并重新紧固相关组件。', 
                        '完成维护后进行了地面试车，未再发现新的渗漏现象。', 
                        '飞机恢复运行状态，记录保留供后续检查使用。', 
                        '起落架在落地过程中右侧收回速度明显低于正常值。', 
                        '飞行员报告指示灯延迟亮起约五秒钟，怀疑液压系统响应缓慢。', 
                        '机务团队在停机坪进行了初步排查。', 
                        '检查结果显示液压油液位偏低，已完成补充。', 
                        '同时对起落架作动筒进行了润滑保养并进行了重复测试。', 
                        '经测试，起落架收放功能恢复正常，未发现进一步异常。']

tokenized_sentences = [
                        ['▁', '发动机', '左', '侧', '在', '飞行', '前', '检查', '中', '发现', '有', '轻微', '的', '机', '油', '渗', '漏', '。'], 
                        ['▁', '技术人员', '确认', '渗', '漏', '区域', '位于', '涡', '轮', '壳', '体', '附近', ',', '并', '进行了', '初步', '清洁', '。'], 
                        ['▁随后', '安排', '进一步', '的', '检查', '以确定', '是否需要', '更换', '密封', '件', '。'], 
                        ['▁在', '机', '务', '人员', '建议', '下', ',', '更换', '了', '左', '侧', '发动机', '的', '油', '封', '并', '重新', '紧', '固', '相关', '组件', '。'], 
                        ['▁', '完成', '维护', '后', '进行了', '地', '面试', '车', ',', '未', '再', '发现', '新的', '渗', '漏', '现象', '。'], 
                        ['▁', '飞机', '恢复', '运行', '状态', ',', '记录', '保留', '供', '后续', '检查', '使用', '。'], 
                        ['▁', '起', '落', '架', '在', '落', '地', '过程中', '右', '侧', '收回', '速度', '明显', '低于', '正常', '值', '。'], 
                        ['▁', '飞行员', '报告', '指示', '灯', '延迟', '亮', '起', '约', '五', '秒钟', ',', '怀疑', '液', '压', '系统', '响应', '缓慢', '。'], 
                        ['▁', '机', '务', '团队', '在', '停', '机', '坪', '进行了', '初步', '排', '查', '。'], 
                        ['▁', '检查', '结果显示', '液', '压', '油', '液', '位', '偏低', ',', '已完成', '补充', '。'], 
                        ['▁同时', '对', '起', '落', '架', '作', '动', '筒', '进行了', '润', '滑', '保养', '并', '进行了', '重复', '测试', '。'], 
                        ['▁经', '测试', ',', '起', '落', '架', '收', '放', '功能', '恢复正常', ',', '未', '发现', '进一步', '异常', '。']
                    ]

def test_expand_to_sentence_level():
    sentencizer = package.sentencizer
    # print(expand_to_sentence_level(rows, sentencizer))
    expanded = expand_to_sentence_level(rows, sentencizer, ignore_empty_paragraph=True, ignore_empty_row=False)
    struct = get_structure(expanded, ignore_empty=True)
    print(expanded)
    assert struct == structure

def test_flatten_sentences():
    # print(flatten_list(expanded_sentences, inner_type=str))
    assert flatten_list(expanded_sentences, inner_type=str) == flattened_sentences

def test_tokenized_sentences():
    tokenizer = package.tokenizer
    sentences_tok =[]
    for sentence in flattened_sentences:
        sentences_tok.append(tokenizer.encode(sentence))
    # print("Sentence tokenized:\n", sentences_tok)
    assert sentences_tok == tokenized_sentences

def test_detokenized_sentences():
    tokenizer = package.tokenizer
    sentences_detok =[]
    for tokens in tokenized_sentences:
        sentences_detok.append(tokenizer.decode(tokens).replace(",", "，")) 
    print("Sentence detokenized:\n", sentences_detok)
    assert sentences_detok == flattened_sentences

def test_get_structure():
    # print(get_structure(expanded_sentences, ignore_empty=True))
    assert get_structure(expanded_sentences, ignore_empty=True) == structure

def test_apply_structure():
    # print(apply_structure(flattened_sentences, structure))
    assert apply_structure(flattened_sentences, structure) == expanded_sentences

def test_text_validator():
    assert is_valid_text("") == False
    assert is_valid_text("   ") == False
    assert is_valid_text("   ! ") == False
    assert is_valid_text("  .; ! ") == False
    assert is_valid_text("0  .; ! ") == True
    assert is_valid_text("0  .; ab ") == True


from . import translator
from . import utils

from .translator.argos_translator import ArgosTranslator
from .translator.package import ArgosPackage
from .translator.package import load_argostranslate_model



from typing import TypedDict, Union, Dict, List, Optional, Callable
from ctranslate2 import  GenerationStepResult
class CT2TranslatorConfig(TypedDict):
    # default configuration based on ctranslate2.Translator, see the documentation for more details https://opennmt.net/CTranslate2/python/ctranslate2.Translator.html#ctranslate2.Translator 
    
    device: str 
    device_index: Union[int, List[int]]
    compute_type: Union[str, Dict[str, str]] 
    inter_threads: int 
    intra_threads: int     
    max_queued_batches: int 
    flash_attention: bool 
    tensor_parallel: bool 
    files: object

class CT2TranslationConfig(TypedDict):
    target_prefix: Optional[List[Optional[List[str]]]] 
    max_batch_size: int 
    batch_type: str 
    asynchronous: bool 
    beam_size: int 
    patience: float 
    num_hypotheses: int 
    length_penalty: float 
    coverage_penalty: float 
    repetition_penalty: float 
    no_repeat_ngram_size: int 
    disable_unk: bool 
    suppress_sequences: Optional[List[List[str]]] 
    end_token: Optional[Union[str, List[str], List[int]]] 
    return_end_token: bool 
    prefix_bias_beta: float 
    max_input_length: int 
    max_decoding_length: int 
    min_decoding_length: int 
    use_vmap: bool 
    return_scores: bool 
    return_logits_vocab: bool 
    return_attention: bool 
    return_alternatives: bool 
    min_alternative_expansion_prob: float 
    sampling_topk: int 
    sampling_topp: float 
    sampling_temperature: float 
    replace_unknowns: bool
    callback: Callable[[GenerationStepResult], bool]


from __future__ import annotations
from typing import List, Union, Dict, TYPE_CHECKING
import os
import psutil
import time

import pyspark
import sentencepiece as spm
import ctranslate2
import stanza
import pandas as pd

from abtranslate.config import CT2TranslatorConfig, CT2TranslationConfig
from abtranslate.config.constants import DEFAULT_CT2_CONFIG, DEFAULT_CT2_TRANSLATION_CONFIG, BATCH_SIZE
from abtranslate.utils.logger import logger
from abtranslate.utils.helper import generate_batch, expand_to_sentence_level, get_structure, apply_structure, restore_expanded_to_rows_level, split_paragraphs, join_paragraphs, flatten_list
from abtranslate.utils.exception import InitializationError

if TYPE_CHECKING:
    from translator.package import ArgosPackage

class ArgosTranslator:
    def __init__(self, package: ArgosPackage, device:str ="cpu", translator_config: CT2TranslatorConfig = DEFAULT_CT2_CONFIG, optimized_config=False):
        self.translator_config = translator_config
        self.device = device
        self.pkg = package
        self.translator = None
        self.using_optimized_config = optimized_config

    def _initialize_models(
        self,
        sample_data
    ) -> ctranslate2.Translator:
        """
        Initialize all required models for translation.
        
        Args:
            compute_type: Computation type for CTranslate2
            inter_threads: Number of inter-threads
            intra_threads: Number of intra-threads
            
        Raises:
            ModelInitializationError: If any model fails to initialize
        """
        if not self.translator:
            try:
                base_translator = ctranslate2.Translator(
                        self.pkg.get_model_path(),
                        self.device,
                        **self.translator_config
                    ) 
                
                if self.using_optimized_config:
                    if len(sample_data) < BATCH_SIZE: # Only if sample data is sufficient, run the translator tuning.  
                        return base_translator
                    self.translator = self.get_optimized_translator(sample_data)
                else:   
                    self.translator = base_translator
            
            except Exception as e:
                logger.error(f"Model initialization error: {e}")
                raise InitializationError(f"Failed to initialize models: {e}")
        
        return self.translator
        
    def _text_preprocessing(self, text: str) -> List[str]:
        tokenizer = self.pkg.tokenizer
        sentences = self.pkg.sentencizer.split_sentences(text) # stanza
        encoded_tokens = [tokenizer.encode(sentence) for sentence in sentences]  # SentencePiece
        return encoded_tokens
    
    def _parse_translation_result(self, ct2_outputs: ctranslate2.TranslationResult) -> List[str]:
        tokenizer = self.pkg.tokenizer
        translated_tokens = [output.hypotheses[0] for output in ct2_outputs]
        translations_detok = [tokenizer.decode(tokens) for tokens in translated_tokens]
        return translations_detok

    def translate(self, input_text:str, translation_config: CT2TranslationConfig = DEFAULT_CT2_TRANSLATION_CONFIG) -> str: 
        """
        Translate a sentence using CTranslate2 and shared SentencePiece model.

        Args:
            input (str): Source sentence to translate
            translator (ctranslate2.Translator): Loaded CTranslate2 model
            sp_model (sentencepiece.SentencePieceProcessor): Loaded shared SentencePiece model

        Returns:
            str: Translated sentence
        """
        translation_result = self.translate_batch([input_text], translation_config) 
        return  translation_result[0]

    def translate_batch(self, text_list: List[str] | pd.Series, translation_config: CT2TranslationConfig = DEFAULT_CT2_TRANSLATION_CONFIG, return_type = List) -> List[str]:
        if isinstance(text_list, pd.Series):
            text_list = text_list.tolist()
            return_type = pd.Series

        translator = self._initialize_models(text_list[:BATCH_SIZE])
        tokenizer = self.pkg.tokenizer
        sentencizer =self.pkg.sentencizer

        expanded_rows = expand_to_sentence_level(text_list, sentencizer, ignore_empty_paragraph=True, ignore_empty_row=False)
        structure = get_structure(expanded_rows, ignore_empty=True)
        sentence_list = flatten_list(expanded_rows, str)
        tokenized_sentences = tokenizer.encode_list(sentence_list)

        if not "max_batch_size" in translation_config.keys():
            translation_config["max_batch_size"] = BATCH_SIZE

        translation_result = translator.translate_batch(
            tokenized_sentences,
            **translation_config
        )
        translated_sentences = self._parse_translation_result(translation_result)
        restored_structure = apply_structure(translated_sentences, structure)
        translated_list = restore_expanded_to_rows_level(restored_structure)
        
        if return_type == pd.Series:
            return pd.Series(translated_list)
        return translated_list

    def get_optimized_translator(self, sample_data: List[str]) -> ctranslate2.Translator:
        best_translator = None
        best_time = float('inf')
        prev_avg = float("inf") 

        translator_config = self.translator_config.copy()
        translator_config["compute_type"] = "int8_float32"
        translation_config = {  "beam_size": 1,
                                "num_hypotheses": 1, 
                                "replace_unknowns": False,}
        
        logical_cpu_count = psutil.cpu_count(logical=False)
        inter_intra_threads_pairs = [(1,                        0), 
                                    (1,                        logical_cpu_count),
                                    (logical_cpu_count//2,      2),
                                    ((logical_cpu_count//2)-1,  2),    
                                    (logical_cpu_count,        0),  
                                    (2,                        logical_cpu_count//2),
                                    (2,                        (logical_cpu_count//2)-1),
                                    (1,                        logical_cpu_count)]
        
        logger.info("Starting CPU allocation benchmark")
        for n_inter_threads, n_intra_threads in inter_intra_threads_pairs:
            translator_config["inter_threads"] = n_inter_threads
            translator_config["intra_threads"] = n_intra_threads
            logger.info(f"Testing translation with inter_threads:{n_inter_threads} intra_threads:{n_intra_threads}")
            translator = ArgosTranslator(
                self.pkg,
                device="cpu",
                translator_config=translator_config,
                optimized_config=False  # Avoid circural logic
                )
            runtimes = []
            for _ in range(3):
                start = time.perf_counter()
                translator.translate_batch(sample_data, translation_config)
                end = time.perf_counter()
                runtimes.append(end - start)

            avg_time = sum(runtimes) / len(runtimes)
            logger.info(f"Translation finished, average time: {avg_time:.4f}s")

            if avg_time < best_time:
                best_time = avg_time
                best_translator = translator.translator
            
            if avg_time > prev_avg:
                patience_count +=1
            else:
                patience_count = 0
            prev_avg = avg_time
        return best_translator